/*		 ITI-FG  Kernel module free public release
 *
 *		     Copyright (C) 1996-2000 GOM mbH.
 *	         Rebenring 33, 38106 Braunschweig, GERMANY
 *		            All rights reserved
 *
 * $Id: itifgMain.c,v 7.0 2000/10/25 09:07:57 cvs Exp $
 *
 */

/*
 * Defines:
 */

/*
 * Includes:
 */

#include "itifgOS.h"	/* os specific stuff */
#include "itifgExt.h"	/* offical driver interface */
#include "itifgReg.h"	/* register definitions ICR Regs */
#include "icpciReg.h"	/* register definitions IC-PCI */
#include "amvsReg.h"	/* register definitions AM-VS */
#include "amdigReg.h"	/* register definitions AM-DIG */
#include "ampcvReg.h"	/* register definitions PCVision */
#include "pcdigReg.h"	/* register definitions PC-DIG */
#include "itifgIntD.h"	/* driver internals */
#include "itifgIntF.h"	/* internal functions */

/*
 * Globals:
 */

int iti_version_code = 0x00070100;	/* means 0.7.1-0 */

int boards;		/* number of attached boards */
 
/* exstime digital camera trigger (first PCDig only) */
int need_update;	/* flag: new exstime value available */
u_char new_exsclk;	/* new prescaler value */
u_long new_exstime;	/* new counter value */
u_long new_midpnt;	/* new midpoint value */

/* main driver structure */
itifg_t *itifg[ITI_BOARDS_MAX];

/*
 * Prototypes:
 */

/*
 * iti_interrupt
 */

void
iti_interrupt (itifg_t *iti)
{
  u_long irqs;
  
  struct timeval daytime;
  
  int percent;

  if (!iti) return;
  
  iti_lock (iti);
  
  irqs = 0UL;
  xcr_test_irqs (iti, &irqs);
  if (!irqs)
    goto done;
  xcr_clear_irqs (iti);
  
  if (!FPGA_OK)
    {
      iti_printn ("run interrupt handler with unconfigured board!\n");
      goto done;
    }

  iti_get_time (&daytime);
  ITI_PRINT6A("> intr: %#lx addr: %#lx %s\n", irqs,
	      iti->get_aqadr (iti),
	      iti->opmode & ITI_MODE_MASK ?
	      "ACQU" : "NONE");	  
  
  /* Bus Master Transfer Done Interrupt */
  if ((IS_ICPCI && (irqs & ICR_BINT)) ||
      (IS_PCDIG && (irqs & ICR_MINT) &&
       (irqs & PCD_BMINT)))
    {
      if (iti->opmode & ITI_MODE_MASK)
	{
	  if (IS_ICPCI)
	    INCR_ACNT(iov);
	  else
	    if (IS_PCDIG)
	      {
		SET_REACHED(iov);
		iti_bma_stop (iti);
	      }
	  ITI_PRINT5A("bm  iov: %d/%d\n", ACNT(iov), DCNT(iov));
	  
	  if (HAS_REACHED(iov))
	    {
	      ZERO_ACNT(iov);
	      INCR_ACNT(blk);
	      ITI_PRINT5A ("bm  blk: %d/%d\n", ACNT(blk), DCNT(blk));
	      if (iti->offset != iti->ob_size /
		  ITI_BLOCK_SIZE - 1 && ACNT(blk) < DCNT(blk) - 1)
		{
		  iti->set_intadr
		    (iti, (((iti->offset + ACNT(blk)) *
			    ITI_BLOCK_SIZE) %
			   iti->ob_size));
		  ITI_PRINT5A ("bm  adr : %#lx\n",
			       iti->get_intadr (iti));
		  iti->enabled = FALSE;
		  if (IS_PCDIG)
		    iti->set_adrien (iti, CR_ENABLE);
		}
	    }
	  else
	    iti_bma_restart (iti);
	  
	  if (HAS_REACHED(blk))
	    {
	      ZERO_ACNT(blk);
	      iti->set_intadr
		(iti, iti->offset * ITI_BLOCK_SIZE);
	      ITI_PRINT5A ("bm  adr : %#lx\n",
			   iti->get_intadr (iti));
	      iti->enabled = FALSE;
	      if (IS_PCDIG)
		iti->set_adrien (iti, CR_ENABLE);
	      
	      iti->acc.frames_transfered++;
	      
	      if (iti->opmode & ITI_MODE_GRAB)
		{
		  struct iti_info_t *frameinfo;
		  
		  if (iti->km_size)
		    {
		      frameinfo = (struct iti_info_t*)
			((caddr_t)iti->km_base +
			 ACNT(cam) * iti->paged_size + iti->raw_size);
		      iti_get_time (&frameinfo->timestamp);
		    }
		  
		  ITI_PRINT4("bm    opmode: GRAB\n");
		  iti_get_time (&daytime);
		  if (!iti_kill_fasync (iti, ITI_CAMERA))
		    {
		      ITI_PRINT4A("bm    status: fasync wakeup "
				  TIME_F "us\n", daytime.tv_usec);
		      iti->acc.frames_get_by_signal++;
		    }
		  else
		    if (!iti_select_wakeup (iti, ITI_CAMERA))
		      {
			ITI_PRINT4A("bm    status: select wakeup "
				    TIME_F "us\n", daytime.tv_usec);
		      }
		    else
		      ITI_PRINT4("Nobody is waiting.\n");
		  iti->opmode &= ~ITI_TIMEOUT;
		  ITI_PRINT4A("bm    opmode: %#x (&~TIME)\n",
			      iti->opmode);
		  iti_add_timer (iti, iti->cam_expire * 2,
				 ITI_CAMERA);
		  ITI_PRINT4("bm    status: timer\n");
		}
	      else
		if (iti->opmode & ITI_MODE_SNAP)
		  {
		    ITI_PRINT4("bm    opmode: SNAP\n");
		    iti_get_time (&iti->timestamp[ACNT(cam)]);
		  }
		else
		  ITI_PRINT4A("bm    opmode: !(GRAB|DONE) %#x\n",
			      iti->opmode);
	      
	      INCR_ACNT(cam);
	      ITI_PRINT5A("bm  cam: %d/%d\n", ACNT(cam), DCNT(cam));
	      iti_bma_stop (iti);
	    }
	  
	  if (HAS_REACHED(cam))
	    {
	      ITI_PRINT5("bm  cam: reached\n");
	      if (iti->opmode & ITI_MODE_GRAB)
		{
		  ITI_PRINT4("bm    opmode: GRAB\n");
			  ZERO_ACNT(cam);
		}
	      else
		if (iti->opmode & ITI_MODE_SNAP)
		  {
		    ITI_PRINT4("bm    opmode: SNAP\n");
		    iti->opmode &= ~ITI_MODE_SNAP;
		    ITI_PRINT4A("bm    opmode: %#x (&~SNAP)\n",
				iti->opmode);
		    iti->opmode |= ITI_SNAP_DONE;
		    iti->opmode &= ~ITI_TIMEOUT;
		    ITI_PRINT4A("bm    opmode: %#x (|DONE&~TIME)\n",
				iti->opmode);
		    iti->started = FALSE;
		    iti_get_time (&daytime);
		    if (!iti_kill_fasync (iti, ITI_CAMERA))
		      {
			ITI_PRINT4A("bm    status: fasync wakeup "
				    TIME_F "us\n", daytime.tv_usec);
			iti->acc.frames_get_by_signal++;
		      }
		    else
		      if (!iti_block_wakeup (iti, ITI_CAMERA))
			{
			  ITI_PRINT4A("bm    status: block wakeup "
				      TIME_F "us\n", daytime.tv_usec);
			}
		      else
			if (!iti_select_wakeup (iti, ITI_CAMERA))
			  {
			    ITI_PRINT4A("bm    status: select wakeup "
					TIME_F "us\n",
					daytime.tv_usec);
			  }
			else
			  ITI_PRINT4("Nobody is waiting.\n");
		  }
		else
		  ITI_PRINT4A("bm    opmode: !(GRAB|DONE) %#x\n",
			      iti->opmode);
	    }
	}
    }
	  
  if (irqs & ICR_MINT)
    {
      /* Verical Blank Interrupts */
      if ((IS_ICPCI && ((irqs & CR_VBINT) ||
			(irqs & CR_OVBINT) ||
			(irqs & CR_EVBINT))) ||
	  (IS_PCDIG && (irqs & PCD_VBINT)))
	{
	  iti->acc.frames_captured++;
	  
	  if ((IS_AMVS || IS_AMPV) && /* analog/interlaced cameras */
	      iti->get_smstat (iti) == CR_ILACE)	  
	    {
	      percent = (iti->get_aqadr (iti) * 100) /
		iti->raw_size;
	      if (percent > 20 && percent < 80)
		iti->swapped = TRUE;
	      else
		iti->swapped = FALSE;
	    }
	  
	  if (iti->opmode & ITI_MODE_MASK)
	    {
	      /* ITI_PRINT4("vb  opmode: ACQU\n"); */
	      
	      if (iti->started)
		{
		  if (!IS_ACNT_ZERO(blk) || DCNT(blk) == 1)
		    {
		      DCNT(iov) = iti->last;
		      ITI_PRINT5A("vb  iov: -/%d\n", DCNT(iov));
		      if (IS_ICPCI)
			{
			  if (DCNT(blk) == 1)
			    iti_bma_start (iti);
			  else
			    iti_bma_restart (iti);
			}
		      else
			if (IS_PCDIG)
			  iti_bma_start (iti);
		    }
		  else
		    ITI_PRINT5("vb  blk: 0/-\n");
		}
	      else
		{
		  ITI_PRINT5("vb  not started yet\n");
		  iti->started = TRUE;
		  if (IS_PCDIG)
		    iti->set_adrien (iti, CR_ENABLE);
		}
	    }
	  else
	    if (iti->opmode & ITI_WAIT_SYNC)
	      {
		struct timeval time;
		
		/* ITI_PRINT4("vb  opmode: SYNC\n"); */
		iti_get_time (&time);
		iti_memcpy (&iti->lasttime, &iti->thistime,
			    sizeof(struct timeval));
		iti_memcpy (&iti->thistime, &time,
			    sizeof(struct timeval));
		
		iti->opmode &= ~ITI_TIMEOUT;
		ITI_PRINT4A("    opmode: %#x (&~TIME)\n",
			    iti->opmode);
		iti_get_time (&daytime);
		if (!iti_block_wakeup (iti, ITI_CAMERA))
		  ITI_PRINT4A("vb    status: block wakeup "
			      TIME_F "us\n", daytime.tv_usec);
		else
		  ITI_PRINT4("Nobody is waiting.\n");
	      }
	  
	  /* update the exsync pulse generator SYNCHRON
	     to new values (set in PCD_SET_ESYNC)  */
	  if (IS_PCDIG && need_update)
	    if (!pcd_update_exsync (iti))
	      need_update = FALSE;
	}
      
      /* Acquisition Address Interrupt */
      if ((IS_ICPCI && (irqs & CR_ADRINT)) ||
	  (IS_PCDIG && (irqs  & PCD_ADRINT)))
	{
	  if (IS_PCDIG)
	    iti->set_adrien (iti, CR_DISABLE);
	  
	  if (iti->opmode & ITI_MODE_MASK)
	    {
	      if (iti->started)
		{
		  /* ITI_PRINT4("aa  opmode: ACQU\n"); */
		  if (IS_ACNT_ZERO(blk))
		    DCNT(iov) = iti->frst;
		  else
		    DCNT(iov) = iti->full;
		  ITI_PRINT5A("aa  iov: -/%d\n", DCNT(iov));
		  
		  if (IS_ACNT_ZERO(blk) || IS_PCDIG)
		    iti_bma_start (iti);
		  else
		    iti_bma_restart (iti);
		}
	      else
		ITI_PRINT5("aa  not started yet\n");
	    }
	}
      
      /* Input Port Interrupt */
      if ((IS_ICPCI && (irqs & CR_INPINT)) ||
	  (IS_PCDIG && (irqs & PCD_PIOINT)))
	iti_iop_interrupt (iti);
      
      /* Acquisition Complete Interrupt */
      if ((IS_ICPCI && irqs & CR_AQINT) ||
	  (IS_PCDIG && irqs & PCD_EOAQINT))
	iti_printn ("Acqusition complete irq!");
      
      /* Trigger Interrupt */
      if ((IS_ICPCI && (irqs & CR_TRIGINT)) ||
	  (IS_PCDIG && (irqs & PCD_TRIG0INT)))
	iti_printn ("Trigger irq!\n");
      
      /* additional PCDig Interrupts */
      if (IS_PCDIG && (irqs & PCD_SWCINT))
	iti_printn ("PCDig software counter irq!\n");
      
      /* additional PCVision Interrupts */
      if (irqs & PV_SOTINT)
	iti_printn ("PCVision strt of trigger irq!\n");
      if (irqs & PV_EOTINT)
	iti_printn ("PCVision end of trig irq!\n");
      if (irqs & PV_EIN0INT)
	iti_printn ("PCVision external irq 0!\n");
      if (irqs & PV_EIN1INT)
	iti_printn ("PCVision external irq 1!\n");
    }
  
  /* Master/Target Abort */
  if (irqs & ICR_MAINT)
    {
      iti_printw ("Master Abort!\n");
      iti_printia ("  dst %#lx, cnt %#lx\n", icr_get_bmdst (iti), 
		   icr_get_bmxc (iti));
    }
  if (irqs & ICR_TAINT)
    {
      iti_printw ("Target Abort!\n");
      iti_printia ("  dst %#lx, cnt %#lx\n", icr_get_bmdst (iti), 
		   icr_get_bmxc (iti));
    }
  
 done:
  iti_unlock (iti);
}

/*
 * iti_cam_timeout
 */

void
iti_cam_timeout (itifg_t *iti)
{
  /* this runs with ALL irq's disabled */

  struct timeval daytime;

  if (!iti) return;

  iti_irq_lock (iti);

  ITI_PRINT0("> timer\n");

  if (!FPGA_OK)
    {
      iti_printn ("run timer handler with unconfigured board!\n");
      goto done;
    }

  if (iti->opmode & ITI_MODE_MASK)
    {
      iti->acc.frames_timedout++;

      ITI_PRINT4("tm    opmode: ACQU\n");

      /* FIXME: one can't disable a running transfer, or the box stucks! */
      /* iti_bma_stop (iti); */
      ZERO_ACNT(cam);
      ZERO_ACNT(blk);
      iti->set_intadr (iti, iti->offset * ITI_BLOCK_SIZE);
      ITI_PRINT5A ("tm  adr : %#lx\n", iti->get_intadr (iti));
      iti->enabled = FALSE;
      if (IS_PCDIG)
	iti->set_adrien (iti, CR_ENABLE);
      ZERO_ACNT(iov);
      iti->started = FALSE;

      if (iti->opmode & ITI_MODE_GRAB)
	{
	  ITI_PRINT4("tm    opmode: GRAB\n");
	  iti_marktq (iti);
	  ITI_PRINT4("tm    status: mark bottom\n");
	}
      else
	if (iti->opmode & ITI_MODE_SNAP)
	  {
	    iti->opmode &= ~ITI_MODE_SNAP;
	    ITI_PRINT4A("tm    opmode: %#x (&~SNAP)\n", iti->opmode);
	    iti->opmode |= ITI_SNAP_DONE | ITI_TIMEOUT;
	    ITI_PRINT4A("tm    opmode: %#x (|DONE|TIME)\n", iti->opmode);
	    iti->started = FALSE;
	    iti_get_time (&daytime);
	    if (!iti_block_wakeup (iti, ITI_CAMERA))
	      ITI_PRINT4A("tm    status: block wakeup "
			  TIME_F "us\n", daytime.tv_usec);
	    else
	      if (!iti_select_wakeup (iti, ITI_CAMERA))
		ITI_PRINT4A("tm    status: select wakeup "
			    TIME_F "us\n", daytime.tv_usec);
	      else
		ITI_PRINT4("Nobody is waiting.\n");
	  }
	else
	  ITI_PRINT4A("tm    opmode: !(GRAB|DONE) %#x\n", iti->opmode);
    }
  else
    if (iti->opmode & ITI_WAIT_SYNC)
      {
	struct timeval time;

	ITI_PRINT4("tm    opmode: SYNC\n");

	/* fill in the framerate detection values */
	iti_get_time (&time);
	iti_memcpy (&iti->lasttime, &iti->thistime,
		    sizeof(struct timeval));
	iti_memcpy (&iti->thistime, &time, sizeof(struct timeval));

	iti->opmode |= ITI_TIMEOUT;
	ITI_PRINT4A("tm    opmode: %#x (|TIME)\n", iti->opmode);
	iti_get_time (&daytime);
	if (!iti_block_wakeup (iti, ITI_CAMERA))
	  ITI_PRINT4A("tm    status: block wakeup "
		      TIME_F "us\n", daytime.tv_usec);
	else
	  if ((IS_AMVS || IS_AMPV) && /* analog/interlaced cameras */
	      iti->get_smstat (iti) == CR_ILACE &&		  
	      iti->get_acqmd (iti) == CR_FREEZE)
	    {
	      iti->set_acqmd (iti, CR_GRAB);
	      iti->opmode &= ~ITI_WAIT_SYNC;
	      iti->opmode |= ITI_MODE_GRAB;
	      ITI_PRINT4A("tm    opmode: %#x (&~SYNC|GRAB)\n",
			  iti->opmode);
	      iti_add_timer (iti, iti->cam_expire * 4, ITI_CAMERA);
	      ITI_PRINT4("tm    status: timer\n");
	    }
	  else
	    ITI_PRINT4("Nobody is waiting.\n");
      }
    else
      ITI_PRINT4A("tm    opmode: NONE %#x\n", iti->opmode);

 done:
  iti_irq_unlock (iti);
}

/*
 * iti_task_queue
 */

void
iti_task_queue (itifg_t *iti)
{
  struct timeval daytime;

  if (!iti) return;

  ITI_PRINT0("> bottom\n");

  if (!FPGA_OK)
    {
      iti_printn ("run bottom half handler with unconfigured board!");
      return;
    }

  if (iti->opmode & ITI_MODE_GRAB)
    {
      ITI_PRINT4("bh    opmode: GRAB\n");

      iti_irq_lock (iti);
      
      iti->opmode |= ITI_TIMEOUT;
      ITI_PRINT4A("bh    opmode: %#x (|TIME)\n", iti->opmode);
      iti_get_time (&daytime);
      if (!iti_block_wakeup (iti, ITI_CAMERA))
	ITI_PRINT4A("tm    status: block wakeup "
		    TIME_F "us\n", daytime.tv_usec);
      else
	if (!iti_select_wakeup (iti, ITI_CAMERA))
	  ITI_PRINT4A("bh    status: select wakeup "
		      TIME_F "us\n", daytime.tv_usec);
	else
	  ITI_PRINT4("Nobody is waiting.\n");
      if ((IS_AMVS || IS_AMPV) && /* analog/interlaced cameras */
	  iti->get_smstat (iti)== CR_ILACE &&		  
	  iti->get_acqmd (iti) == CR_GRAB)
	{
	  iti->set_acqmd (iti, CR_FREEZE);
	  iti->opmode &= ~ITI_MODE_GRAB;
	  iti->opmode |= ITI_WAIT_SYNC;
	  ITI_PRINT4A("bh    opmode: %#x (&~GRAB|SYNC)\n", iti->opmode);
	  iti_add_timer (iti, iti->cam_expire * 4, ITI_CAMERA);
	  ITI_PRINT4("bh    status: timer\n");
	}

      iti_irq_unlock (iti);

      /* fill the buffer with a grey image */
      if (iti->km_size)
	iti_clear (iti, iti->km_base, iti->km_size, ITI_CLEAR_GREY, FALSE);

    }
  else
    ITI_PRINT4A("bh    opmode: !GRAB %#x\n", iti->opmode);
}

/*
 * iti_open
 */

int
iti_open (iti_file_t *file, iti_dev_t dev)
{
  int error, board, funct;
  itifg_t *iti;

  if ((error = iti_decode_devno (dev, &board, &funct)))
    return error; /* only ENXIO */
  iti = itifg[board];

  ITI_PRINT0A("> iti_open %d: lock %d\n", board, iti->cam_read_lock);

  /* PCDig is not configured, but we have to write it */
  if (!FPGA_OK)
    goto open_done;

  /* the io-port stuff is handled in a seperate file */
  if (IS_PORT)
    {
      if (IS_REV_B || IS_PCDIG)
	{
	  error = iti_iop_open (iti, file);
	  if (error)
	    goto open_error;
	  else
	    goto open_done;
	}
      else
	{
	  error = -ENXIO;
	  goto open_error;
	}
    }

  /* for write only, nothing more is to do */
  if (((F_FLAGS(file) & O_ACCMODE) == O_WRONLY))
    goto open_done;
  else
    {
      /* O_RDWR - O_RDONLY: test for lock */
      if (F_FLAGS(file) & O_EXCL)
	{
	  if (iti->cam_read_lock)
	    return -EBUSY;
	}
      else
	if (iti->cam_read_lock < 0)
	  return -EBUSY;
    }

  /* first read open -> initialize */
  if (!iti->cam_read_lock)
    {
      /* reset statistics */
      iti_irq_lock (iti);
      iti->acc.frames_captured = 0UL;
      iti->acc.frames_transfered = 0UL;
      iti->acc.frames_timedout = 0UL;
      iti->acc.frames_get_by_blocked = 0UL;
      iti->acc.frames_get_by_signal = 0UL;
      iti->acc.frames_get_by_select = 0UL;
      iti_irq_unlock (iti);

      /* reset the select frame marker */
      iti->frames_selected = 0;
      /* reset the fpga write stuff */
      iti->exo_buffer = NULL;
      iti->exo_count = 0;
      /* reset frame rate detection times */
      iti_memset (&iti->thistime, 0x00, sizeof(struct timeval));
      iti_memset (&iti->lasttime, 0x00, sizeof(struct timeval));
      /* reset the aio call structure */
      iti->aio_buffer = NULL;
      iti->aio_count = 0;

      /* set default operation mode and timeout */
      iti->opmode = ITI_MODE_NONE;
      ITI_PRINT4("    opmode: NONE\n");
      /* iti->expire = (float)HZ / ITI_INIT_FPS; FIXME */

      /* enable main interrupts (if not already enabled) */
      if (!iti->iop_read_lock)
	if ((error = iti_enable_irq (iti, ITI_EIRQ_MAIN_OPEN)))
	  goto open_error;

      /* enable grabber interrupts */
      if ((error = iti_enable_irq (iti, ITI_EIRQ_DATA_OPEN)))
	goto open_error;

      /* reset the hardware decimation */
      if (iti->set_hdec (iti, CR_DEC_BY_1) ||
	  iti->set_vdec (iti, CR_DEC_BY_1))
	{
	  iti_error (iti, "iti->vcr.set_[hv]");
	  error = -EIO;
	  goto open_error;
	}

      /* start image acqusition */
      if (iti->set_acqmd (iti, CR_GRAB))
	{
	  iti_error (iti, "iti->vcr.set_acqmd");
	  error = -EIO;
	  goto open_error;
	}
    }

  /* lock device now */
  if (((F_FLAGS(file) & O_ACCMODE) == O_RDONLY) ||
      ((F_FLAGS(file) & O_ACCMODE) == O_RDWR))
    {
      if (F_FLAGS(file) & O_EXCL)
	iti->cam_read_lock = -1;
      else
	iti->cam_read_lock++;
    }

 open_done:
  ITI_PRINT0A("< iti_open %d: lock %d\n", board, iti->cam_read_lock);
  return OK;

 open_error:
  ITI_PRINT0A("< iti_open %d: error %d\n", board, error);
  return error;
}

/*
 * iti_close
 */

int
iti_close (iti_file_t *file, iti_inode_t *inode, iti_dev_t dev)
{
  int error, board, funct;
  itifg_t *iti;

  if ((error = iti_decode_devno (dev, &board, &funct)))
    return error; /* only ENXIO */
  iti = itifg[board];

  ITI_PRINT0A("> iti_close %d: lock %d\n", board, iti->cam_read_lock);

  /* PCDig isn't still configured, perhaps write has failed */
  if (!FPGA_OK)
    goto close_done;

  /* the io-port stuff is handled in a seperate file */
  if (IS_PORT)
    {
      if (IS_REV_B || IS_PCDIG)
	{
	  error = iti_iop_close (iti, file);
	  if (error)
	    goto close_error;
	  else
	    goto close_done;
	}
      else
	{
	  error = -ENXIO;
	  goto close_error;
	}
    }

  /* for write only, nothing more is to do */
  if (((F_FLAGS(file) & O_ACCMODE) == O_WRONLY))
    goto close_done;

  /* if last process, close the device */
  if (iti->cam_read_lock == 1 || iti->cam_read_lock == -1)
    {
      /* stop image acquisition */
      if (iti->set_acqmd (iti, CR_FREEZE))
	{
	  iti_error (iti, "iti->vcr.set_acqmd");
	  error = -EIO;
	  goto close_error;
	}

      /* FIXME: one can't disable a running transfer, or the box stucks! */
      /* iti_bma_stop (iti); */

      /* disable grabber interrupts */
      if (!iti->iop_read_lock)
	iti_disable_irq (iti, ITI_DIRQ_DATA_CLOSE);

      /* disable main interrupts (if not used by others) */
      if (!iti->iop_read_lock)
	iti_disable_irq (iti, ITI_DIRQ_MAIN_CLOSE);

      /* reset operation mode */
      iti->opmode = ITI_MODE_NONE;
      ITI_PRINT4("    opmode: NONE\n");

      /* cleanup signal type and target */
      iti_init_fasync (iti, file, inode, -1, FALSE);

      /* clean any pending timers */
      iti_del_timer (iti, ITI_CAMERA);

      /* release the grab buffer */
      if (iti->km_size)
	if ((error = iti_unmap (file, dev, iti->km_base, iti->km_size)))
	  return error;
    }

  /* unlock devices now */
  if (iti->cam_read_lock < 0)	/* O_EXCL */
    iti->cam_read_lock = 0;
  else
    if (iti->cam_read_lock)
      iti->cam_read_lock--;
    else
      {
	error = -EBADF;
	goto close_error;
      }

 close_done:
  ITI_PRINT0A("< iti_close %d: lock %d\n", board, iti->cam_read_lock);
  return OK;

 close_error:
  ITI_PRINT0A("< iti_close %d: error %d\n", board, error);
  return error;
}

/* magic sequences (used for read test purposes) */
/* should NEVER found in valid (8,16,24bit) images !!?? */
#define ITI_MEM_TEST	0x0F00FFAA
#define ITI_UMB_TEST	0xF0FF0055

/*
 * iti_read
 */

iti_ssize_t
iti_read (iti_file_t *file, iti_dev_t dev, char *buffer, size_t count)
{
  int error, board, funct, frame, frames;
  itifg_t *iti;

  caddr_t base;
  u_long page;
  size_t size;

  if ((error = iti_decode_devno (dev, &board, &funct)))
    return error; /* only ENXIO */
  iti = itifg[board];

  ITI_PRINT0A("> iti_read %d: at 0x%p\n", board, buffer);
  ITI_PRINT0A("    " SIZE_F " byte(s)\n", count);

   /* if some of the FPGA's are not configured, we can't read anything */
  if (!FPGA_OK)
    {
      error = -ENXIO;
      goto read_error;
    }

  /* is the buffer pointer valid at all? */
  if (buffer == NULL)
    {
      error = -EINVAL;
      goto read_error;
    }

  /* all io-port stuff is handled by separate functions */
  if (IS_PORT)
    {
      count = iti_iop_read (iti, file, buffer, count);
      if (count < 0)
	{
	  error = count;
	  goto read_error;
	}
      else
	goto read_done;
    }

  /* nothing to do (is this the right behavior?) */
  frames = 0;
  if (!count)
    goto read_done;

  /* is the buffer big enough to hold at least one frame? */
  if (count < iti->paged_size)
    {
      error = -EINVAL;
      goto read_error;
    }
  
  /* calculate the frame count */
  frames = count / iti->paged_size;
  ITI_PRINT1A("    frames: %d\n", frames);

  /* is this the completion of an async i/o? */
  ITI_PRINT1A("    aio_buffer: 0x%p\n", iti->aio_buffer);
  ITI_PRINT1A("    aio_count: " SIZE_F "\n", iti->aio_count);
  if (iti->aio_buffer)
    {
      if (buffer != iti->aio_buffer)
	{
	  error =  -EBUSY;
	  goto read_error;
	}
      if (iti->aio_count)
	{
	  if (count != iti->aio_count)
	    {
	      error = -EBUSY;
	      goto read_error;
	    }
	  ITI_PRINT1("    aio stop call\n");
	  goto aio_return;
	}
    }

  /* has any other driver activation successfully finished? */
  if (iti->opmode & ITI_WAIT_MASK)
    {
      ITI_PRINT4("    opmode: USED\n");
      /* force sane state */
      iti->aio_buffer = NULL;
      iti->aio_count = 0;
      error = -EBUSY;
      goto read_error;
    }
  ITI_PRINT4("    opmode: !USED\n");

  /* fix interlace 'odd/even frame swapped' problems */ 
  if ((IS_AMVS || IS_AMPV) && iti->swapped &&
      iti->get_smstat (iti) == CR_ILACE)
    {
      /* stop image acquisition (synchron) */
      iti->set_acqmd (iti, CR_FREEZE);
      iti->opmode |= ITI_WAIT_SYNC;
      ITI_PRINT4A("    opmode: %#x (|SYNC)\n", iti->opmode);
      ITI_PRINT4("    status: sleep\n");
      if ((error = iti_sleep (iti, iti->cam_expire, ITI_CAMERA)))
	goto read_error;
      iti->opmode &= ~ITI_WAIT_SYNC;
      ITI_PRINT4A("    opmode: %#x (&~SYNC)\n", iti->opmode);
      /* start image acqusition (synchron) */
      iti->set_acqmd (iti, CR_GRAB);
      iti->opmode |= ITI_WAIT_SYNC;
      ITI_PRINT4A("    opmode: %#x (|SYNC)\n", iti->opmode);
      ITI_PRINT4("    status: sleep\n");
      if ((error = iti_sleep (iti, iti->cam_expire, ITI_CAMERA)))
	goto read_error;
      iti->opmode &= ~ITI_WAIT_SYNC;
      ITI_PRINT4A("    opmode: %#x (&~SYNC)\n", iti->opmode);
      iti->swapped = FALSE;
      iti_printn ("analog camera odd/even frames swapped!\n");
      /* force sane state */
      iti->aio_buffer = NULL;
      iti->aio_count = 0;
      error = -EIO;
      goto read_error;
    }

  /* set source and destination scatter gather tables */
  if ((error = iti_src_init (iti)))
    goto read_error; /* always ok */
   
  /* if the user hasn't lock the memory yet (by mlock), we have to do it */
  if ((error = iti_area_locked (buffer)))
    {
      if (error != -EFAULT)
	return error;
      else
	{
	  base = buffer;
	  size = count;
	  
	  /* lock page by page */
	  while (size > 0) 
	    {
	      if ((error = iti_virt_to_phys (base, &page, NULL, TRUE)))
	      return error;
	      ITI_PRINT3A("    v:0x%p->p:%#lx\n", base, page);
	      if ((error = iti_lock_page (page, TRUE)))
		return error;
	    base += PAGE_SIZE;
	    size -= PAGE_SIZE;
	    }
	}
    }
  
  if ((error = iti_dst_init (iti, buffer,
			     frames * iti->paged_size, TRUE)))
    {
      /* force sane state */
      iti->aio_buffer = NULL;
      iti->aio_count = 0;
      goto read_error; /* ENOMEM, EFAULT, EINVAL */
    }

   /* write the magic sequence to ONE src frame (at the end) */
  {
    void *src_addr;

    src_addr = (caddr_t)iti->ob_base +
      (iti->raw_size % iti->ob_size) - sizeof (u_long);
    iti_writel (ITI_MEM_TEST, ((u_long*)(src_addr)));
  }
  /* write the magic sequence and EVERY dst frame (at the end) */
  for (frame = 0; frame < frames; frame++)
    {
      void *dst_addr;
      off_t dst_off;
      u_long magic = ITI_UMB_TEST;
      
      dst_addr = buffer;
      dst_off = frame * iti->paged_size;
      if ((error = iti_copyout (((void*)((caddr_t)dst_addr + dst_off)),
				&magic, sizeof(u_long))))
	goto read_error;
    }

  /* now acquire some frames */
  iti_irq_lock (iti);
  ZERO_ACNT(cam);
  ZERO_ACNT(blk);
  iti->set_intadr (iti, iti->offset * ITI_BLOCK_SIZE);
  iti->enabled = FALSE;
  if (IS_PCDIG)
    iti->set_adrien (iti, CR_ENABLE);
  /* if an acq. is already running, delay this until the next vsync */
  if ((IS_AMVS || IS_AMPV) && iti->get_aqadr) /* FIXME */
    iti->started = FALSE;
  else
    iti->started = TRUE;
  ZERO_ACNT(iov);
  iti->opmode |= ITI_MODE_SNAP;
  ITI_PRINT4A("    opmode: %#x (|SNAP)\n", iti->opmode);
  iti_irq_unlock (iti);

   /* set timer and wait on queue or return from async request start */
  if (F_FLAGS(file) & O_NONBLOCK)
    {
      iti->aio_buffer = buffer;
      iti->aio_count = count;
      ITI_PRINT1("    aio start call\n");
      ITI_PRINT1A("    aio_buffer: 0x%p\n", iti->aio_buffer);
      ITI_PRINT1A("    aio_count: " SIZE_F "\n", iti->aio_count);
      if ((error = iti_add_timer (iti,
				  iti->cam_expire * (frames + 1) +
				  iti->cam_expire / 2,
				  ITI_CAMERA)))
	goto read_error; /* no error */
      ITI_PRINT4("    status: timer\n");
      count = 0; /* till now no byte is read */
      goto read_done;
    }
  else
    {
      ITI_PRINT1("    no aio call\n");
      ITI_PRINT4("    status: sleep\n");
      if ((error = iti_sleep (iti, iti->cam_expire * (frames + 1) +
			      iti->cam_expire / 2,
			      ITI_CAMERA)))
	return error; /* ERESTARTSYS */
      iti->acc.frames_get_by_blocked++;
    }
  
 aio_return:
  
  /* has the image acquisition really finished */
  if (!(iti->opmode & ITI_SNAP_DONE))
    {
      ITI_PRINT4("    opmode: !DONE\n");
      error = -EBUSY;
      goto read_error;
    }
  ITI_PRINT4("    opmode: DONE\n");
  iti->opmode &= ~ITI_SNAP_DONE;
  ITI_PRINT4A("    opmode: %#x (&~SNAP)\n", iti->opmode);  
  
  /* now reset the vectors */
  iti->aio_buffer = NULL;
  iti->aio_count = 0;

  /* outside this read any initialized vectors are invalid! */
  iti_clear_sglist (iti);

  if (iti_area_locked (buffer) == -EFAULT)
    {
      base = buffer;
      size = count;
      
      /* unlock page by page */
      while (size > 0) 
	{
	  if ((error = iti_virt_to_phys (base, &page, NULL, TRUE)))
	    return error;
	  if ((error = iti_unlock_page (page, TRUE)))
	    return error;
	  base += PAGE_SIZE;
	  size -= PAGE_SIZE;
	}
    }

  /* this flag is set, when no frame is read after the expire period */ 
  if (iti->opmode & ITI_TIMEOUT)
    {
      ITI_PRINT4("    opmode: TIME\n");
      iti->opmode &= ~ITI_TIMEOUT;
      ITI_PRINT4A("    opmode: %#x (&~TIME)\n", iti->opmode);  
      iti_clear (iti, buffer, count, ITI_CLEAR_GREY, TRUE);
      iti_printn ("timeout while reading frame!\n");
      error = -ENODATA;
      goto read_error;
    }
  ITI_PRINT4("    opmode: !TIME\n");

  /* test for the magic sequence in ONE src image */
  {
    void *src_addr;
    
    src_addr = (caddr_t)iti->ob_base + (iti->raw_size % iti->ob_size) -
      sizeof (u_long);

    if (iti_readl (((u_long*)(src_addr))) == ITI_MEM_TEST)
      {
	iti_printn ("frame not correctly acquired!\n");
	error = -EIO;
	goto read_error;
      }
  }

  /* test for the magic sequence in ALL dst images and set the frame time */
  for (frame = 0; frame < frames; frame++)
    {
      struct iti_info_t *frameinfo;
      void *dst_addr;
      off_t dst_off;
      u_long magic;
      
      dst_addr = buffer;
      dst_off = frame * iti->paged_size;
      if ((error = iti_copyin (&magic, (void*)((caddr_t)dst_addr + dst_off),
			       sizeof (u_long))))
	goto read_error;
      if (magic == ITI_UMB_TEST)
	{
	  iti_printna ("frame %d not correctly transfered!\n", frame);
	  error = -EIO;
	  goto read_error;
	}
      
      frameinfo = (struct iti_info_t*)(buffer +
				       frame * iti->paged_size +
				       iti->raw_size);
      if ((error = iti_copyout (&frameinfo->timestamp,
				&iti->timestamp[frame],
				sizeof(struct timeval))))
	goto read_error;
    }
  
read_done:
  ITI_PRINT0A("< iti_read %d: " SIZE_F " byte(s)\n", board, count);
  return count;
  
read_error:
  ITI_PRINT0A("< iti_read %d: error %d\n", board, error);
  return error;
}

/*
 * iti_write
 */

iti_ssize_t
iti_write (iti_file_t *file, iti_dev_t dev, const char *buffer, size_t count)
{
  int error, board, funct;
  itifg_t *iti;

  if ((error = iti_decode_devno (dev, &board, &funct)))
    return error; /* only ENXIO */
  iti = itifg[board];
  
  ITI_PRINT0A("> iti_write %d: at 0x%p\n", board, buffer);
  ITI_PRINT0A("    " SIZE_F " byte(s)\n", count);

  /* if we try 'no module', we actually mean the io-ports */
  if (IS_PORT)
    {
      count =  iti_iop_write (iti, file, buffer, count);
      if (count < 0)
	{
	  error = count;
	  goto write_error;
	}
      else
	goto write_done;
    }

  /* has any other driver activation successfully finished? */
  if (iti->opmode & ITI_WAIT_MASK)
    {
      ITI_PRINT4("    opmode: USED\n");
      error = -EBUSY;
      goto write_error;
    }
  ITI_PRINT4("    opmode: !USED\n");

  /* only on the PCDig we can write the FPGA code */
  if (!IS_PCDIG)
    {
      error = -ENXIO;
      goto write_error;
    }

  /* alloc the kernel data buffer */
  if (iti->exo_buffer == NULL)
    {
      iti->exo_buffer = iti_malloc (PCD_CONFIG_BYTES, ITI_PAGED);
      if (iti->exo_buffer == NULL)
	{
	  error = -EAGAIN;
	  goto write_error;
	}
      ITI_PRINT3A("    iti->exo_buffer: 0x%p\n", iti->exo_buffer);
      iti->exo_count = 0;
    }

  /* cut the count if too much data and get the data */
  if (iti->exo_count + count > PCD_CONFIG_BYTES)
    count = PCD_CONFIG_BYTES - iti->exo_count;
  if ((error = iti_copyin ((iti->exo_buffer + iti->exo_count),
			   buffer, count)))
    goto write_error;
  iti->exo_count += count;

  /* if not enought data to start download, return for now */
  if (iti->exo_count < PCD_CONFIG_BYTES)
    {
      ITI_PRINT1A("    exo count incomplete: " SIZE_F "\n",
		  iti->exo_count);
      goto write_done;
    }
  ITI_PRINT1A("    exo count complete: " SIZE_F "\n",
	      iti->exo_count);

  /* if this is not the first download */
  if (FPGA_OK)
    {
      /* stop image acquisition */
      if (iti->set_acqmd (iti, CR_FREEZE))
	{
	  iti_error (iti, "iti->vcr.set_acqmd");
	  error = -EIO;
	  goto write_error;
	}
      /* disable busmaster requests */
      iti_bma_stop (iti);
      /* disable main interrupts */
      if ((error = iti_disable_irq (iti, ITI_DIRQ_MAIN_CLOSE)))
	goto write_error;
    }
  /* else: FIXME - save the current register configuration */

  /* from now the hardware is not existent */
  iti->fpga_ok = FALSE;
  
  /* download config file to fpga */
  if ((error = pcd_load_fpga (iti, iti->exo_buffer)) < 0)
    {
      iti_printw ("Load of PC-DIG FPGA's failed!\n");
      ITI_PRINT3A("    iti->exo_buffer: 0x%p\n", iti->exo_buffer);
      iti_free (iti->exo_buffer, ITI_PAGED);
      iti->exo_buffer = NULL;
      iti->exo_count = 0;
      goto write_error; /* EIO */
    }

  /* free the now unused buffer */
  ITI_PRINT3A("    iti->exo_buffer: 0x%p\n", iti->exo_buffer);
  iti_free (iti->exo_buffer, ITI_PAGED);
  iti->exo_buffer = NULL;
  iti->exo_count = 0;

  /* initialize master clock generator */
  {
    u_long clkvalue;
    
    if (pcd_calc_clkval (iti, PCD_CLKVAL_SYSTEM, &clkvalue))
      {
	iti_error (iti, "pcd_calc_clkval");
	error = -EIO;
	goto write_error;
      }
    if (pcd_write_clkval (iti, 'B', clkvalue))
      {
	iti_error (iti, "pcd_write_clkval");
	error = -EIO;
	goto write_error;
      }
  }
  if (pcd_set_synmuxb (iti, PCD_SYNMUX_PROG))
    {
      iti_error (iti, "pcd_set_synmuxb");
      error = -EIO;
      goto write_error;
    }

  /* clear on-board SGRAM and FIFO */
  if (pcd_init_sginit (iti))
    {
      iti_error (iti, "pcd_init_sginit");
      error = -EIO;
      goto write_error;
    }
  if (pcd_init_fifoinit (iti))
    {
      iti_error (iti, "pcd_init_fifoinit");
      error = -EIO;
      goto write_error;
    }
  
  /* from now any registers are available again */
  iti->fpga_ok = TRUE;

  /* init the register set to defined values */
  if ((error = iti_cr_init (iti)))
    {
      iti_printw ("Init. of Control Regs. failed.\n");  
      goto write_error;
    }

  /* write init values to last dma table entry??? */
  if (pcd_set_dmaacc (iti, PCD_ENABLE))
    {
      iti_error (iti, "pcd_set_dmaacc");
      error = -EIO;
      goto write_error;
    }
  iti_writel (PCD_INIT_AGCSTS,
	      &iti->pcd_dma_table[PCD_NUM_DMA_ENTRYS - 1].addr);
  iti_writel (PCD_INIT_AINT, 
	      &iti->pcd_dma_table[PCD_NUM_DMA_ENTRYS - 1].xcnt);
  if (pcd_set_dmaacc (iti, PCD_DISABLE))
    {
      iti_error (iti, "pcd_set_dmaacc");
      error = -EIO;
      goto write_error;
    }
  if ((error = pcd_viam_startup (iti)))
    {
      iti_printw ("Startup of PC-DIG Regs. failed.\n");
      goto write_error;
    }
  if ((error = pcd_viam_init (iti)))
    {
      iti_printw ("Init. of PC-DIG Regs. failed.\n");
      goto write_error;
    }
  if ((error = iti_fmt_init (iti)))
    {
      iti_printw ("Format initialize failed.\n");
      goto write_error;
    }

  /* reenable the right irq's */
  if ((error = iti_enable_irq (iti, ITI_EIRQ_MAIN_OPEN)))
    return error;
  if (iti->cam_read_lock)
    if ((error = iti_enable_irq (iti, ITI_EIRQ_DATA_OPEN)))
      return error;
  if (iti->iop_read_lock)
    if ((error = iti_enable_irq (iti, ITI_EIRQ_PORT_OPEN)))
      return error;

  /* start image acqusition */
  if (iti->set_acqmd (iti, CR_GRAB))
    {
      iti_error (iti, "iti->vcr.set_acqmd");
      error = -EIO;
      goto write_error;
    }

write_done:
  ITI_PRINT0A("< iti_write %d: " SIZE_F " byte(s)\n", board, count);
  return count;
  
write_error:
  ITI_PRINT0A("< iti_write %d: error %d\n", board, error);
  return error;
}  

/*
 * iti_lseek
 */

iti_loff_t
iti_lseek (iti_file_t *file, iti_dev_t dev, iti_loff_t offset, int mode)
{
  int error, board, funct, frame;
  itifg_t *iti;

  if ((error = iti_decode_devno (dev, &board, &funct)))
    return error; /* only ENXIO */
  iti = itifg[board];
  
  ITI_PRINT0A("> iti_lseek %d: to " LOFF_F ", mode %d\n", board, offset, mode);

  /* no seek is on the io-ports possible */
  if (IS_PORT)
    {
      error = -ESPIPE;
      goto lseek_error;
    }

  /* this fuction not actually seeks, it returns only the offset */
  switch (mode)
    {
    case 0: /* SEEK_SET */
      error = -ESPIPE;
      goto lseek_error;
      break;
    case 1: /* SEEK_CUR */
      if (offset)
	{
	  error = -ESPIPE;
	  goto lseek_error;
	}
      break;
    case 2: /* SEEK_END */
      error = -ESPIPE;
      goto lseek_error;
      break;
    }

  /* ... and is only useful (for now) in grab mode */
  if (!(iti->opmode & ITI_MODE_GRAB))
    {
      ITI_PRINT4("    opmode: !GRAB\n");
      offset = 0;
      goto lseek_done;
    }
  ITI_PRINT4("    opmode: GRAB\n");

  /* the current frame -1 is the latest whole frame available */ 
  frame = ACNT(cam);
  if (frame)
    frame--;
  else
    frame = DCNT(cam) - 1;
  ITI_PRINT1A("    frame: %d\n", frame);

  /* only whole frames are interesting */
  offset = frame * iti->paged_size;

  goto lseek_done;

lseek_done:
  ITI_PRINT0A("< iti_lseek %d: is " LOFF_F "\n", board, offset);
  return offset;
  
lseek_error:
  ITI_PRINT0A("< iti_lseek %d: error %d\n", board, error);
  return error;
}

/*
 * iti_mmap
 */

int
iti_mmap (iti_file_t *file, iti_dev_t dev, caddr_t base, size_t size)
{
  int error, board, funct;
  itifg_t *iti;

  caddr_t src, dst;
  u_long page;
  size_t len;

  if ((error = iti_decode_devno (dev, &board, &funct)))
    return error; /* only ENXIO */
  iti = itifg[board];
  
  ITI_PRINT0A("> iti_mmap %d: at 0x%p\n", board, base);
  ITI_PRINT0A("    " SIZE_F " byte(s)\n", size);

  /* is the size page aligned */
  if (size % PAGE_SIZE)
    {
      error = -EFAULT;
      goto mmap_error;
    }

  /* if the PCDig is not configured, we can't mmap it */
  if (!FPGA_OK)
    {
      error = -ENXIO;
      goto mmap_error;
    }

  /* can never mmap the io-ports */
  if (IS_PORT)
    {
      error = -ENXIO;
      goto mmap_error;
    }

  /* if it is already mapped.. */
  if (size)
    if (iti->km_size)
      {
	error = -EBUSY;
	goto mmap_error;
      }

  /* alloc some virtual memory space */
  src = iti_malloc (size, ITI_PAGED);
  if (!src)
    {
      error = -EAGAIN;
      goto mmap_error;
    }

  iti->km_base = src;
  ITI_PRINT3A("    iti->buf->km_base: 0x%p\n", iti->km_base);
  iti->km_size = size;  

  /* make every page present, lock and remap it */
  dst = base;
  len = size;
  while (len > 0) 
    {
      if ((error = iti_virt_to_phys (src, &page, NULL, FALSE)))
	return error;
      ITI_PRINT3A("    0x%p v:0x%p->p:%#lx\n", dst, src, page);
      if ((error = iti_lock_page (page, FALSE)))
	return error;
      if ((error = iti_remap_page (dst, page)))
	return error;
      src += PAGE_SIZE;
      dst += PAGE_SIZE;
      len -= PAGE_SIZE;
    }

  /* wipe out the alloced memory */
  iti_clear (iti, iti->km_base, iti->km_size, ITI_CLEAR_BLACK, FALSE);
  
  goto mmap_done;

mmap_done:
  ITI_PRINT0A("< iti_mmap %d: OK\n", board);
  return OK;
  
mmap_error:
  ITI_PRINT0A("< iti_mmap %d: error %d\n", board, error);
  return error;
}

/*
 * iti_unmap
 */

int
iti_unmap (iti_file_t *file, iti_dev_t dev, caddr_t base, size_t size)
{
  int error, board, funct;
  itifg_t *iti;

  u_long page;

  if ((error = iti_decode_devno (dev, &board, &funct)))
    return error; /* only ENXIO */
  iti = itifg[board];
    
  ITI_PRINT0A("> iti_unmmap %d: at 0x%p\n", board, base);
  ITI_PRINT0A("    " SIZE_F " byte(s)\n", size);

  /* if noting is mapped */
  if (!size)
    {
      error = -ENOMEM;
      goto unmap_error;
    }
  
  /* clear the transfer vector list, before freeing the buffer */
  iti_clear_sglist (iti);

  /* unlock page by page */
  while (size > 0) 
    {
      if ((error = iti_virt_to_phys (base, &page, NULL, FALSE)))
	return error;
      if ((error = iti_unlock_page (page, FALSE)))
	return error;
      base += PAGE_SIZE;
      size -= PAGE_SIZE;
    }

  /* free the buffer */
  ITI_PRINT3A("    iti->buf->km_base: 0x%p\n", iti->km_base);
  iti_free (iti->km_base, ITI_PAGED);
  iti->km_base = NULL;
  iti->km_size = 0;
  
  goto unmap_done;

unmap_done:
  ITI_PRINT0A("< iti_unmap %d: OK\n", board);
  return OK;
  
unmap_error:
  ITI_PRINT0A("< iti_unmap %d: error %d\n", board, error);
  return error;
}
  
/*
 * iti_select
 */

int
iti_select (iti_file_t *file, iti_dev_t dev, int arg_type)
{
  int error, board, funct, ret_type;
  itifg_t *iti;

  if ((error = iti_decode_devno (dev, &board, &funct)))
    return error; /* only ENXIO */
  iti = itifg[board];
    
  ITI_PRINT0A("> iti_select %d: type %#x\n", board, arg_type);

  /* PCDig is not configured, we can't select on it */
  if (!FPGA_OK)
    {
      error = -ENXIO;
      goto select_error;
    }

  /* can't yet select on io-ports */
  if (IS_PORT)
    {
      error = -ENXIO;
      goto select_error;
    }

  /* reset the return value */
  ret_type = 0;
  
  /* input event */
  if (arg_type & ITI_SEL_IN)
    {
      if (iti->opmode & ITI_MODE_GRAB)
	{
	  ITI_PRINT4("    opmode: GRAB\n");
	  ITI_PRINT1A ("     select: %ld\n", iti->frames_selected);
	  ITI_PRINT1A ("     transfer: %ld\n",
		       iti->acc.frames_transfered);
	  if (iti->acc.frames_transfered > iti->frames_selected)
	    {
	      iti->acc.frames_get_by_select++;
	      iti->frames_selected = iti->acc.frames_transfered;
	      ret_type |= ITI_SEL_IN;
	    }
	}
      else
	if (iti->opmode & ITI_SNAP_DONE)
	  {
	    ITI_PRINT4("    opmode: DONE\n");
	    iti->acc.frames_get_by_select++;
	    ret_type |= ITI_SEL_IN;
	  }
	else
	  ITI_PRINT4A("    opmode: !(GRAB|DONE) %#x\n", iti->opmode);
    }
  
  /* exeption event */ 
  if (arg_type & ITI_SEL_EX)
    {
      if (iti->opmode & ITI_TIMEOUT)
	{
	  ITI_PRINT4("    opmode: TIME\n");
	  if (iti->opmode & ITI_MODE_GRAB)
	    {
	      ITI_PRINT4("    opmode: TIME\n");
	      /* reset the TIMEOUT flag */
	      iti->opmode &= ~ITI_TIMEOUT;
	      ITI_PRINT4A("    opmode: %#x (&~TIME)\n", iti->opmode);
	      ret_type |= ITI_SEL_EX;
	    }
	  else
	    if (iti->opmode & ITI_SNAP_DONE)
	      {
		ITI_PRINT4("    opmode: DONE\n");
		/* Don't reset the TIMEOUT flag (is done by aio_return) */
		ret_type |= ITI_SEL_EX;
	      }
	    else
	      if (iti->opmode & ITI_WAIT_SYNC)
		{
		  ITI_PRINT4("    opmode: SYNC\n");
		  if (IS_AMVS || IS_AMPV) /* analog boards */
		    ret_type |= ITI_SEL_EX;
		}
	}
      else
	ITI_PRINT4("    opmode: !TIME\n");
    }
  
  goto select_done;

select_done:
  ITI_PRINT0A("< iti_select %d: %#x\n", board, ret_type);
  return ret_type;
  
select_error:
  ITI_PRINT0A("< iti_select %d: error %d\n", board, error);
  return error;
}

/*
 * iti_fasync
 */

int
iti_fasync (iti_file_t *file, iti_inode_t *inode, iti_dev_t dev,
	    int desc, int mode)
{
  int error, board, funct;
  itifg_t *iti;

  if ((error = iti_decode_devno (dev, &board, &funct)))
    return error; /* only ENXIO */
  iti = itifg[board];
    
  ITI_PRINT0A("> iti_fasync %d: mode %#x\n", board, mode);

  /* PCDig is not configured, we can't async events from it */
  if (!FPGA_OK)
    {
      error = -ENXIO;
      goto fasync_error;
    }
  
  /* can't yet send signals on io-port events */
  if (IS_PORT)
    {
      error = -ENXIO;
      goto fasync_error;
    }

  /* completely handled from the os part */
  error = iti_init_fasync (iti, file, inode, desc, mode);
  if (error)
    goto fasync_error;
  else
    goto fasync_done;

fasync_done:
  ITI_PRINT0A("< iti_fasync %d: OK\n", board);
  return OK;
  
fasync_error:
  ITI_PRINT0A("< iti_fasync %d: error %d\n", board, error);
  return error;
}

